@@ -5,9 +5,9 @@ import json
 import logging
 import os
 import time
+from collections import deque
 from dataclasses import asdict, astuple, dataclass
 from decimal import Decimal
-from collections import deque
 
 from mysql.connector import MySQLConnection, connect
 from mysql.connector.pooling import MySQLConnectionPool
@@ -45,6 +45,18 @@ def init_logger(name: str | None = None) -> logging.Logger:
     return logger
 
 
+def hash_dict(d):
+    return frozenset(d.items())
+
+
+def find_missing_in_b(a, b):
+    b_hashes = {hash_dict(d) for d in b}
+
+    missing_in_b = [d for d in a if hash_dict(d) not in b_hashes]
+
+    return missing_in_b
+
+
 def get_table_rows_number(con: MySQLConnection, database: str, table: str) -> int:
     with con.cursor() as cur:
         cur.execute("SELECT table_rows FROM information_schema.tables WHERE table_schema = %s AND table_name = %s", (database, table))
@@ -62,7 +74,7 @@ def get_table_structure(con: MySQLConnection, database: str, table: str) -> list
 
 
 def get_table_keys(con: MySQLConnection, database: str, table: str) -> list[tuple[str, str]]:
-    query_table_keys_statement = """
+    operation = """
         SELECT tis.index_name, titc.constraint_type, tic.column_name, tic.data_type
         FROM information_schema.table_constraints titc
         JOIN information_schema.statistics tis ON titc.table_schema = tis.table_schema AND titc.table_name = tis.table_name AND titc.constraint_name = tis.index_name
@@ -73,7 +85,7 @@ def get_table_keys(con: MySQLConnection, database: str, table: str) -> list[tupl
     """
 
     with con.cursor() as cur:
-        cur.execute(query_table_keys_statement, (database, table))
+        cur.execute(operation, (database, table))
         rows = cur.fetchall()
 
     keys = [(row[2], row[3]) for row in rows if row[1] == "PRIMARY KEY"] or [(row[2], row[3]) for row in rows if row[1] == "UNIQUE" and row[0] == rows[0][0]]
@@ -139,30 +151,6 @@ class MysqlTableCompare:
         if self.limit_size % self.fetch_size != 0:
             raise Exception("The remainder of the fetch_size divided by the shard_size must be equal to 0.")
 
-    # batch_id, source_key_vals, diff_rows
-    def processing_result(self, batch_id: int, cts: list[ComparisonTask]):
-        cts.sort(key=lambda task: task.batch_id)
-
-        while cts and cts[0].batch_id == batch_id:
-            _, source_rows, diff_rows = astuple(cts.pop(0))
-
-            self.processed_rows_number += len(source_rows)
-
-            if diff_rows:
-                self.write_different(diff_rows)
-                self.different_rows_number += len(diff_rows)
-
-            if self.processed_rows_number % 10000 == 0 or diff_rows:
-                self.checkpoint.processed = self.processed_rows_number
-                self.checkpoint.different = self.different_rows_number
-                self.checkpoint.row = extract_keyvals(source_rows[-1], self.source_table_keys)
-                self.write_checkpoint()
-                self.logger.info(f"checkpoint: {self.checkpoint}")
-
-            batch_id += 1
-
-        return batch_id
-
     def write_different(self, rows: list[dict]):
         with open(self.different_file, "a", encoding="utf8") as f:
             for row in rows:
@@ -187,24 +175,7 @@ class MysqlTableCompare:
                     ckpt.row[coln] = Decimal(ckpt.row[coln])
         return ckpt
 
-    def get_table_rows_by_keys(self, con: MySQLConnectionPool, database: str, table: str, table_keys_rows: list[dict]) -> list[dict]:
-        cols = [key[0] for key in self.source_table_keys]
-        placeholders = ", ".join(["%s"] * len(cols))
-
-        in_clause = ", ".join([f"({placeholders})" for _ in table_keys_rows])
-
-        params = [val for row in table_keys_rows for val in (row[col] for col in cols)]
-
-        formatted_cols = [f"`{c}`" for c in cols]
-        _stmt = f"SELECT * FROM {database}.{table} WHERE ({', '.join(formatted_cols)}) IN ({in_clause}) ORDER BY {', '.join(formatted_cols)}"
-
-        with con.get_connection() as con:
-            con: MySQLConnection
-            with con.cursor(dictionary=True, buffered=True) as cur:
-                cur.execute(_stmt, tuple(params))
-                return cur.fetchall()
-
-    def get_full_table_keys_order(self, ckpt_row: dict = None):
+    def get_queery_full_table_statement_params(self, ckpt_row: dict = None):
         _keyval = ckpt_row
         # select * from where 1 = 1 and ((a > xxx) or (a = xxx and b > yyy) or (a = xxx and b = yyy and c > zzz)) order by a,b,c limit checksize
         _key_colns = ", ".join([f"`{col[0]}`" for col in self.source_table_keys])
@@ -227,110 +198,66 @@ class MysqlTableCompare:
         statement_with_condition = f"SELECT * FROM {self.src_database}.{self.src_table} {where_clause} ORDER BY {_key_colns} LIMIT {self.limit_size}"
         statement_without_condition = f"SELECT * FROM {self.src_database}.{self.src_table} ORDER BY {_key_colns} LIMIT {self.limit_size}"
 
-        with connect(**self.source_dsn) as source_con:
-            source_con: MySQLConnection
-            _start_time_0 = time.time()
+        _params: list = []
+        if _keyval:
+            for end_idx in range(len(self.source_table_keys)):
+                for i, (column_name, _) in enumerate(self.source_table_keys[: end_idx + 1]):
+                    _params.append(_keyval[column_name])
 
-            while True:
-                _start_time_1 = time.time()
-                _params: list = []
-                _rows_queue = deque()
+        statement = statement_with_condition if _params else statement_without_condition
 
-                if _keyval:
-                    for end_idx in range(len(self.source_table_keys)):
-                        for i, (column_name, _) in enumerate(self.source_table_keys[: end_idx + 1]):
-                            _params.append(_keyval[column_name])
+        return statement, _params
 
-                statement = statement_with_condition if _params else statement_without_condition
+    def query_rows(self, con: MySQLConnectionPool, query_statement: str, query_params) -> list[dict]:
+        with con.get_connection() as con:
+            con: MySQLConnection
+            with con.cursor(dictionary=True, buffered=True) as cur:
+                cur.execute(query_statement, tuple(query_params))
+                return cur.fetchall()
 
-                self.logger.debug(f"full table query - sql: {statement}, params: {_params}, {_keyval}")
+    def compare_full_table(self):
+        _keyval = self.checkpoint.row
+        print(f"keyval: {_keyval}")
 
-                with source_con.cursor(dictionary=True, buffered=True) as cur:
-                    cur.execute(statement, params=_params)
+        while True:
+            query_statement, query_params = self.get_queery_full_table_statement_params(_keyval)
+            print(f"query {query_statement}, params {query_statement}")
 
-                    for row in cur.fetchall():
-                        _rows_queue.append(row)
+            source_rows = []
+            target_rows = []
 
-                _rows_cnt = len(_rows_queue)
+            _start1 = time.time()
 
-                if _rows_cnt >= 1:
-                    self.logger.debug(f"full table query - fetch once query: {get_elapsed_time(_start_time_1, 2)}s, fetch rows count: {_rows_cnt}.")
+            with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
+                future_source = executor.submit(self.query_rows, self.source_conpool, query_statement, query_params)
+                future_target = executor.submit(self.query_rows, self.target_conpool, query_statement, query_params)
 
-                    while _rows_queue:
-                        rows = []
-                        for _ in range(min(self.fetch_size, len(_rows_queue))):
-                            rows.append(_rows_queue.popleft())
+                source_rows = future_source.result()
+                target_rows = future_target.result()
 
-                        last_row = rows[-1]
-                        _keyval = extract_keyvals(last_row, self.source_table_keys)
-                        yield rows
-                else:
-                    self.logger.debug(f"full table query - fetch all rows: {get_elapsed_time(_start_time_0, 2)}s.")
-                    return
+            print(f"query source & target rows, source rows: {len(source_rows)}, target rows {len(target_rows)} {get_elapsed_time(_start1, 2)}s.")
 
-    def compare_rows_by_keys(self, ct: ComparisonTask):
-        _start_time = time.time()
-
-        keyval_list = []
-        for row in ct.source_rows:
-            keyval_list.append(extract_keyvals(row, self.source_table_keys))
-
-        target_rows = self.get_table_rows_by_keys(self.target_conpool, self.dst_database, self.dst_table, keyval_list)
-
-        ct.different += list(itertools.filterfalse(lambda x: x in target_rows, ct.source_rows))
-
-        self.logger.debug(
-            f"compare rows success - batch[{ct.batch_id}] - elapsed time {get_elapsed_time(_start_time,2)}s, rows: {len(ct.source_rows)}, different: {len(ct.different)}."
-        )
-
-        return ct
-
-    def compare_full_table(self, checkpoint_row):
-        _batch_id = 1
-        # batch_id, source_key_vals, diff_rows = _tasks.pop(0)
-        _comparison_tasks_container: list[ComparisonTask] = []
-
-        fulltasks = enumerate(self.get_full_table_keys_order(checkpoint_row), start=1)
-
-        with concurrent.futures.ThreadPoolExecutor(max_workers=self.parallel) as executor:
-
-            def generator_futures(n):
-                try:
-                    return [
-                        executor.submit(self.compare_rows_by_keys, ComparisonTask(batch_id, source_row, []))
-                        for batch_id, source_row in itertools.islice(fulltasks, n)
-                    ]
-                except Exception as e:
-                    raise Exception(f"full table query failure - error: {e}.")
-
-            futures = generator_futures(self.parallel)
-
-            while futures:
-                done, no_done = concurrent.futures.wait(futures, return_when=concurrent.futures.FIRST_COMPLETED)
-
-                for fut in done:
-                    _comparison_tasks_container.append(fut.result())
-                futures = [fut for fut in futures if fut not in done]
-                _batch_id = self.processing_result(_batch_id, _comparison_tasks_container)
-
-                futures += generator_futures(len(done))
-
-                self.logger.debug(
-                    (
-                        f"compare progress - done: {len(done)}, no_done: {len(no_done)}, "
-                        f"processed: {self.processed_rows_number}, "
-                        f"progress: {round(self.processed_rows_number / self.source_table_rows_number * 100, 1)}%, "
-                        f"different: {self.different_rows_number}, "
-                        f"total rows: {self.source_table_rows_number}."
-                    )
-                )
+            if len(source_rows) == 0:
+                break
+            else:
+                _keyval = extract_keyvals(source_rows[-1], self.source_table_keys)
+                _start = time.time()
+                diff_rows = find_missing_in_b(source_rows, target_rows)
+                print(f"find_missing_in_b {get_elapsed_time(_start, 2)}s.")
+                if len(diff_rows) >= 1:
+                    self.write_different(diff_rows)
+                self.checkpoint.row = _keyval
+                self.write_checkpoint()
+                print(f"Missing elements in b: {len(diff_rows)}")
+                print(f"keyval: {_keyval}")
 
     def run(self) -> None:
-        if os.path.exists(self.done_file):
-            return
-
         self.logger = init_logger(self.compare_name)
 
+        if os.path.exists(self.done_file):
+            self.logger.info(f"The comparison is complete and '{self.done_file}' exists.")
+            return
+
         self.logger.info("init check.")
 
         with connect(**self.source_dsn) as source_con, connect(**self.target_dsn) as target_con:
@@ -364,7 +291,7 @@ class MysqlTableCompare:
 
         self.logger.info("target table connect pool create.")
         self.target_conpool = MySQLConnectionPool(pool_name="target_conpool", pool_size=self.parallel, **self.target_dsn)
-        self.compare_full_table(self.checkpoint.row)  # main
+        self.compare_full_table()  # main
 
         self.logger.info(
             f"compare completed, processed rows: {self.processed_rows_number}, different: {self.different_rows_number},  elapsed time: {get_elapsed_time(self.start_timestamp, 2)}s."
@@ -374,3 +301,9 @@ class MysqlTableCompare:
 
         if os.path.exists(self.checkpoint_file):
             os.remove(self.checkpoint_file)
+
+
+if __name__ == "__main__":
+    source_dsn = {"host": "192.168.161.94", "user": "cy", "password": "cy"}
+    target_dsn = {"host": "192.168.161.94", "user": "cy", "password": "cy"}
+    MysqlTableCompare(source_dsn, target_dsn, "merchant_center_ouku_v1", "v3_wh_to_whg", "merchant_center_ouku_v1", "v3_wh_to_whg", 1, 200, 200).run()
